# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: confidence/flags/resolver/v1/types.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    3,
    '',
    'confidence/flags/resolver/v1/types.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(confidence/flags/resolver/v1/types.proto\x12\x1c\x63onfidence.flags.resolver.v1\"e\n\x03Sdk\x12\x31\n\x02id\x18\x01 \x01(\x0e\x32#.confidence.flags.resolver.v1.SdkIdH\x00\x12\x13\n\tcustom_id\x18\x02 \x01(\tH\x00\x12\x0f\n\x07version\x18\x03 \x01(\tB\x05\n\x03sdk*\xa4\x03\n\rResolveReason\x12\x1e\n\x1aRESOLVE_REASON_UNSPECIFIED\x10\x00\x12\x18\n\x14RESOLVE_REASON_MATCH\x10\x01\x12#\n\x1fRESOLVE_REASON_NO_SEGMENT_MATCH\x10\x02\x12)\n!RESOLVE_REASON_NO_TREATMENT_MATCH\x10\x03\x1a\x02\x08\x01\x12 \n\x1cRESOLVE_REASON_FLAG_ARCHIVED\x10\x04\x12&\n\"RESOLVE_REASON_TARGETING_KEY_ERROR\x10\x05\x12\x18\n\x14RESOLVE_REASON_ERROR\x10\x06\x12.\n*RESOLVE_REASON_UNRECOGNIZED_TARGETING_RULE\x10\x07\x12\x30\n,RESOLVE_REASON_MATERIALIZATION_NOT_SUPPORTED\x10\x08\x12 \n\x1cRESOLVE_REASON_TYPE_MISMATCH\x10\t\x12!\n\x1dRESOLVE_REASON_FLAG_NOT_FOUND\x10\n*\xa3\x05\n\x05SdkId\x12\x16\n\x12SDK_ID_UNSPECIFIED\x10\x00\x12\x18\n\x14SDK_ID_JAVA_PROVIDER\x10\x01\x12\x1a\n\x16SDK_ID_KOTLIN_PROVIDER\x10\x02\x12\x19\n\x15SDK_ID_SWIFT_PROVIDER\x10\x03\x12\x1a\n\x16SDK_ID_JS_WEB_PROVIDER\x10\x04\x12\x1d\n\x19SDK_ID_JS_SERVER_PROVIDER\x10\x05\x12\x1a\n\x16SDK_ID_PYTHON_PROVIDER\x10\x06\x12\x16\n\x12SDK_ID_GO_PROVIDER\x10\x07\x12\x18\n\x14SDK_ID_RUBY_PROVIDER\x10\x08\x12\x18\n\x14SDK_ID_RUST_PROVIDER\x10\t\x12\x1a\n\x16SDK_ID_JAVA_CONFIDENCE\x10\n\x12\x1c\n\x18SDK_ID_KOTLIN_CONFIDENCE\x10\x0b\x12\x1b\n\x17SDK_ID_SWIFT_CONFIDENCE\x10\x0c\x12\x18\n\x14SDK_ID_JS_CONFIDENCE\x10\r\x12\x1c\n\x18SDK_ID_PYTHON_CONFIDENCE\x10\x0e\x12\x18\n\x14SDK_ID_GO_CONFIDENCE\x10\x0f\x12\x1a\n\x16SDK_ID_RUST_CONFIDENCE\x10\x10\x12!\n\x1dSDK_ID_FLUTTER_IOS_CONFIDENCE\x10\x11\x12%\n!SDK_ID_FLUTTER_ANDROID_CONFIDENCE\x10\x12\x12\x1c\n\x18SDK_ID_DOTNET_CONFIDENCE\x10\x13\x12\x1c\n\x18SDK_ID_GO_LOCAL_PROVIDER\x10\x14\x12\x1e\n\x1aSDK_ID_JAVA_LOCAL_PROVIDER\x10\x15\x12#\n\x1fSDK_ID_JS_LOCAL_SERVER_PROVIDER\x10\x16\x42\x9f\x01\n,com.spotify.confidence.sdk.flags.resolver.v1B\nTypesProtoP\x01Zagithub.com/spotify/confidence-resolver/openfeature-provider/go/confidence/internal/proto/resolverb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'confidence.flags.resolver.v1.types_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n,com.spotify.confidence.sdk.flags.resolver.v1B\nTypesProtoP\001Zagithub.com/spotify/confidence-resolver/openfeature-provider/go/confidence/internal/proto/resolver'
  _globals['_RESOLVEREASON'].values_by_name["RESOLVE_REASON_NO_TREATMENT_MATCH"]._loaded_options = None
  _globals['_RESOLVEREASON'].values_by_name["RESOLVE_REASON_NO_TREATMENT_MATCH"]._serialized_options = b'\010\001'
  _globals['_RESOLVEREASON']._serialized_start=178
  _globals['_RESOLVEREASON']._serialized_end=598
  _globals['_SDKID']._serialized_start=601
  _globals['_SDKID']._serialized_end=1276
  _globals['_SDK']._serialized_start=74
  _globals['_SDK']._serialized_end=175
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: confidence/wasm/messages.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'confidence/wasm/messages.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..flags.resolver.v1 import types_pb2 as confidence_dot_flags_dot_resolver_dot_v1_dot_types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1e\x63onfidence/wasm/messages.proto\x12\x0f\x63onfidence.wasm\x1a(confidence/flags/resolver/v1/types.proto\"\x06\n\x04Void\"l\n\x17SetResolverStateRequest\x12\r\n\x05state\x18\x01 \x01(\x0c\x12\x12\n\naccount_id\x18\x02 \x01(\t\x12.\n\x03sdk\x18\x03 \x01(\x0b\x32!.confidence.flags.resolver.v1.Sdk\"\x17\n\x07Request\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"5\n\x08Response\x12\x0e\n\x04\x64\x61ta\x18\x01 \x01(\x0cH\x00\x12\x0f\n\x05\x65rror\x18\x02 \x01(\tH\x00\x42\x08\n\x06result\"-\n\x19PrometheusSnapshotRequest\x12\x10\n\x08instance\x18\x01 \x01(\t\"*\n\x1aPrometheusSnapshotResponse\x12\x0c\n\x04text\x18\x01 \x01(\tB\x8c\x01\n\x1f\x63om.spotify.confidence.sdk.wasmB\x08MessagesP\x00Z]github.com/spotify/confidence-resolver/openfeature-provider/go/confidence/internal/proto/wasmb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'confidence.wasm.messages_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\037com.spotify.confidence.sdk.wasmB\010MessagesP\000Z]github.com/spotify/confidence-resolver/openfeature-provider/go/confidence/internal/proto/wasm'
  _globals['_VOID']._serialized_start=93
  _globals['_VOID']._serialized_end=99
  _globals['_SETRESOLVERSTATEREQUEST']._serialized_start=101
  _globals['_SETRESOLVERSTATEREQUEST']._serialized_end=209
  _globals['_REQUEST']._serialized_start=211
  _globals['_REQUEST']._serialized_end=234
  _globals['_RESPONSE']._serialized_start=236
  _globals['_RESPONSE']._serialized_end=289
  _globals['_PROMETHEUSSNAPSHOTREQUEST']._serialized_start=291
  _globals['_PROMETHEUSSNAPSHOTREQUEST']._serialized_end=336
  _globals['_PROMETHEUSSNAPSHOTRESPONSE']._serialized_start=338
  _globals['_PROMETHEUSSNAPSHOTRESPONSE']._serialized_end=380
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: confidence/flags/resolver/v1/api.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'confidence/flags/resolver/v1/api.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from . import types_pb2 as confidence_dot_flags_dot_resolver_dot_v1_dot_types__pb2
from ...types.v1 import types_pb2 as confidence_dot_flags_dot_types_dot_v1_dot_types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&confidence/flags/resolver/v1/api.proto\x12\x1c\x63onfidence.flags.resolver.v1\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a(confidence/flags/resolver/v1/types.proto\x1a%confidence/flags/types/v1/types.proto\"\xaf\x01\n\x13ResolveFlagsRequest\x12\r\n\x05\x66lags\x18\x01 \x03(\t\x12\x33\n\x12\x65valuation_context\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x15\n\rclient_secret\x18\x03 \x01(\t\x12\r\n\x05\x61pply\x18\x04 \x01(\x08\x12.\n\x03sdk\x18\x05 \x01(\x0b\x32!.confidence.flags.resolver.v1.Sdk\"\x85\x01\n\x14ResolveFlagsResponse\x12\x42\n\x0eresolved_flags\x18\x01 \x03(\x0b\x32*.confidence.flags.resolver.v1.ResolvedFlag\x12\x15\n\rresolve_token\x18\x02 \x01(\x0c\x12\x12\n\nresolve_id\x18\x03 \x01(\t\"\xda\x01\n\x11\x41pplyFlagsRequest\x12\x38\n\x05\x66lags\x18\x01 \x03(\x0b\x32).confidence.flags.resolver.v1.AppliedFlag\x12\x15\n\rclient_secret\x18\x02 \x01(\t\x12\x15\n\rresolve_token\x18\x03 \x01(\x0c\x12-\n\tsend_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\x03sdk\x18\x05 \x01(\x0b\x32!.confidence.flags.resolver.v1.Sdk\"\x14\n\x12\x41pplyFlagsResponse\"K\n\x0b\x41ppliedFlag\x12\x0c\n\x04\x66lag\x18\x01 \x01(\t\x12.\n\napply_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\xf5\x01\n\x0cResolvedFlag\x12\x0c\n\x04\x66lag\x18\x01 \x01(\t\x12\x0f\n\x07variant\x18\x02 \x01(\t\x12&\n\x05value\x18\x03 \x01(\x0b\x32\x17.google.protobuf.Struct\x12K\n\x0b\x66lag_schema\x18\x04 \x01(\x0b\x32\x36.confidence.flags.types.v1.FlagSchema.StructFlagSchema\x12;\n\x06reason\x18\x05 \x01(\x0e\x32+.confidence.flags.resolver.v1.ResolveReason\x12\x14\n\x0cshould_apply\x18\x06 \x01(\x08\x32\x8e\x01\n\x13\x46lagResolverService\x12w\n\x0cResolveFlags\x12\x31.confidence.flags.resolver.v1.ResolveFlagsRequest\x1a\x32.confidence.flags.resolver.v1.ResolveFlagsResponse\"\x00\x42\x9d\x01\n,com.spotify.confidence.sdk.flags.resolver.v1B\x08\x41piProtoP\x01Zagithub.com/spotify/confidence-resolver/openfeature-provider/go/confidence/internal/proto/resolverb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'confidence.flags.resolver.v1.api_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n,com.spotify.confidence.sdk.flags.resolver.v1B\010ApiProtoP\001Zagithub.com/spotify/confidence-resolver/openfeature-provider/go/confidence/internal/proto/resolver'
  _globals['_RESOLVEFLAGSREQUEST']._serialized_start=217
  _globals['_RESOLVEFLAGSREQUEST']._serialized_end=392
  _globals['_RESOLVEFLAGSRESPONSE']._serialized_start=395
  _globals['_RESOLVEFLAGSRESPONSE']._serialized_end=528
  _globals['_APPLYFLAGSREQUEST']._serialized_start=531
  _globals['_APPLYFLAGSREQUEST']._serialized_end=749
  _globals['_APPLYFLAGSRESPONSE']._serialized_start=751
  _globals['_APPLYFLAGSRESPONSE']._serialized_end=771
  _globals['_APPLIEDFLAG']._serialized_start=773
  _globals['_APPLIEDFLAG']._serialized_end=848
  _globals['_RESOLVEDFLAG']._serialized_start=851
  _globals['_RESOLVEDFLAG']._serialized_end=1096
  _globals['_FLAGRESOLVERSERVICE']._serialized_start=1099
  _globals['_FLAGRESOLVERSERVICE']._serialized_end=1241
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: confidence/flags/types/v1/types.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'confidence/flags/types/v1/types.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%confidence/flags/types/v1/types.proto\x12\x19\x63onfidence.flags.types.v1\"\xce\x06\n\nFlagSchema\x12O\n\rstruct_schema\x18\x01 \x01(\x0b\x32\x36.confidence.flags.types.v1.FlagSchema.StructFlagSchemaH\x00\x12K\n\x0blist_schema\x18\x02 \x01(\x0b\x32\x34.confidence.flags.types.v1.FlagSchema.ListFlagSchemaH\x00\x12I\n\nint_schema\x18\x03 \x01(\x0b\x32\x33.confidence.flags.types.v1.FlagSchema.IntFlagSchemaH\x00\x12O\n\rdouble_schema\x18\x04 \x01(\x0b\x32\x36.confidence.flags.types.v1.FlagSchema.DoubleFlagSchemaH\x00\x12O\n\rstring_schema\x18\x05 \x01(\x0b\x32\x36.confidence.flags.types.v1.FlagSchema.StringFlagSchemaH\x00\x12K\n\x0b\x62ool_schema\x18\x06 \x01(\x0b\x32\x34.confidence.flags.types.v1.FlagSchema.BoolFlagSchemaH\x00\x1a\xbc\x01\n\x10StructFlagSchema\x12R\n\x06schema\x18\x01 \x03(\x0b\x32\x42.confidence.flags.types.v1.FlagSchema.StructFlagSchema.SchemaEntry\x1aT\n\x0bSchemaEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x34\n\x05value\x18\x02 \x01(\x0b\x32%.confidence.flags.types.v1.FlagSchema:\x02\x38\x01\x1a\x12\n\x10\x44oubleFlagSchema\x1a\x0f\n\rIntFlagSchema\x1a\x12\n\x10StringFlagSchema\x1a\x10\n\x0e\x42oolFlagSchema\x1aO\n\x0eListFlagSchema\x12=\n\x0e\x65lement_schema\x18\x01 \x01(\x0b\x32%.confidence.flags.types.v1.FlagSchemaB\r\n\x0bschema_typeB\x99\x01\n)com.spotify.confidence.sdk.flags.types.v1B\nTypesProtoP\x01Z^github.com/spotify/confidence-resolver/openfeature-provider/go/confidence/internal/proto/typesb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'confidence.flags.types.v1.types_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n)com.spotify.confidence.sdk.flags.types.v1B\nTypesProtoP\001Z^github.com/spotify/confidence-resolver/openfeature-provider/go/confidence/internal/proto/types'
  _globals['_FLAGSCHEMA_STRUCTFLAGSCHEMA_SCHEMAENTRY']._loaded_options = None
  _globals['_FLAGSCHEMA_STRUCTFLAGSCHEMA_SCHEMAENTRY']._serialized_options = b'8\001'
  _globals['_FLAGSCHEMA']._serialized_start=69
  _globals['_FLAGSCHEMA']._serialized_end=915
  _globals['_FLAGSCHEMA_STRUCTFLAGSCHEMA']._serialized_start=556
  _globals['_FLAGSCHEMA_STRUCTFLAGSCHEMA']._serialized_end=744
  _globals['_FLAGSCHEMA_STRUCTFLAGSCHEMA_SCHEMAENTRY']._serialized_start=660
  _globals['_FLAGSCHEMA_STRUCTFLAGSCHEMA_SCHEMAENTRY']._serialized_end=744
  _globals['_FLAGSCHEMA_DOUBLEFLAGSCHEMA']._serialized_start=746
  _globals['_FLAGSCHEMA_DOUBLEFLAGSCHEMA']._serialized_end=764
  _globals['_FLAGSCHEMA_INTFLAGSCHEMA']._serialized_start=766
  _globals['_FLAGSCHEMA_INTFLAGSCHEMA']._serialized_end=781
  _globals['_FLAGSCHEMA_STRINGFLAGSCHEMA']._serialized_start=783
  _globals['_FLAGSCHEMA_STRINGFLAGSCHEMA']._serialized_end=801
  _globals['_FLAGSCHEMA_BOOLFLAGSCHEMA']._serialized_start=803
  _globals['_FLAGSCHEMA_BOOLFLAGSCHEMA']._serialized_end=819
  _globals['_FLAGSCHEMA_LISTFLAGSCHEMA']._serialized_start=821
  _globals['_FLAGSCHEMA_LISTFLAGSCHEMA']._serialized_end=900
# @@protoc_insertion_point(module_scope)



"""CDN state fetcher for Confidence OpenFeature provider.

This module provides functionality to fetch resolver state from the Confidence CDN.
"""

import hashlib
import logging
from typing import Optional, Tuple

import httpx

from confidence.proto.confidence.wasm.messages_pb2 import (
    SetResolverStateRequest,
)

logger = logging.getLogger(__name__)

CDN_BASE_URL = "https://confidence-resolver-state-cdn.spotifycdn.com"


class StateFetcherError(Exception):
    """Exception raised when state fetching fails."""

    pass


class StateFetcher:
    """Fetches and caches resolver state from the Confidence CDN.

    The StateFetcher uses ETag-based caching to efficiently poll for state updates.
    It builds the CDN URL from a SHA256 hash of the client secret.

    Example:
        fetcher = StateFetcher("your-client-secret")
        state, account_id = fetcher.fetch()
    """

    def __init__(
        self,
        client_secret: str,
        http_client: Optional[httpx.Client] = None,
    ) -> None:
        """Initialize the StateFetcher.

        Args:
            client_secret: The Confidence client secret used to build the CDN URL.
            http_client: Optional httpx.Client for custom HTTP configuration or testing.
        """
        self._client_secret = client_secret
        self._http_client = http_client
        self._owns_client = http_client is None

        # Cached state
        self._state: Optional[bytes] = None
        self._account_id: Optional[str] = None
        self._etag: Optional[str] = None

        # Build CDN URL from SHA256 hash of client secret
        hash_hex = hashlib.sha256(client_secret.encode()).hexdigest()
        self._cdn_url = f"{CDN_BASE_URL}/{hash_hex}"

    @property
    def state(self) -> Optional[bytes]:
        """Return the current cached resolver state."""
        return self._state

    @property
    def account_id(self) -> Optional[str]:
        """Return the current cached account ID."""
        return self._account_id

    def _get_client(self) -> httpx.Client:
        """Get or create the HTTP client."""
        if self._http_client is not None:
            return self._http_client
        # Create a new client for this request
        return httpx.Client(timeout=30.0)

    def fetch(self) -> Tuple[bytes, str, bool]:
        """Fetch the resolver state from the CDN.

        This method fetches the latest resolver state from the Confidence CDN.
        It uses ETag-based caching to avoid re-downloading unchanged state.

        Returns:
            A tuple of (state_bytes, account_id, changed) where changed indicates
            whether the state was updated (False means 304 Not Modified).

        Raises:
            StateFetcherError: If the HTTP request fails or returns an error status.
        """
        client = self._get_client()
        should_close = self._http_client is None

        try:
            headers = {}
            if self._etag is not None:
                headers["If-None-Match"] = self._etag

            response = client.get(self._cdn_url, headers=headers)

            # Handle 304 Not Modified - return cached state
            if response.status_code == 304:
                if self._state is not None and self._account_id is not None:
                    logger.debug("State not modified (304), using cached state")
                    return self._state, self._account_id, False
                raise StateFetcherError(
                    "Received 304 Not Modified but no cached state available"
                )

            # Handle error responses
            if response.status_code != 200:
                raise StateFetcherError(
                    f"Failed to fetch state: HTTP {response.status_code}"
                )

            # Parse the protobuf response
            state_request = SetResolverStateRequest()
            state_request.ParseFromString(response.content)

            # Cache the state, account ID, and ETag
            self._state = state_request.state
            self._account_id = state_request.account_id
            self._etag = response.headers.get("ETag")

            logger.info(
                "Loaded resolver state for account=%s, etag=%s",
                self._account_id,
                self._etag,
            )

            return self._state, self._account_id, True

        finally:
            if should_close:
                client.close()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: confidence/flags/resolver/v1/internal_api.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'confidence/flags/resolver/v1/internal_api.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from . import types_pb2 as confidence_dot_flags_dot_resolver_dot_v1_dot_types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n/confidence/flags/resolver/v1/internal_api.proto\x12\x1c\x63onfidence.flags.resolver.v1\x1a\x1fgoogle/protobuf/timestamp.proto\x1a(confidence/flags/resolver/v1/types.proto\"\xb6\x02\n\x14WriteFlagLogsRequest\x12\x41\n\rflag_assigned\x18\x01 \x03(\x0b\x32*.confidence.flags.resolver.v1.FlagAssigned\x12\x43\n\x0etelemetry_data\x18\x02 \x01(\x0b\x32+.confidence.flags.resolver.v1.TelemetryData\x12L\n\x13\x63lient_resolve_info\x18\x03 \x03(\x0b\x32/.confidence.flags.resolver.v1.ClientResolveInfo\x12H\n\x11\x66lag_resolve_info\x18\x04 \x03(\x0b\x32-.confidence.flags.resolver.v1.FlagResolveInfo\"\x17\n\x15WriteFlagLogsResponse\"?\n\rTelemetryData\x12.\n\x03sdk\x18\x02 \x01(\x0b\x32!.confidence.flags.resolver.v1.Sdk\"g\n\nClientInfo\x12\x0e\n\x06\x63lient\x18\x01 \x01(\t\x12\x19\n\x11\x63lient_credential\x18\x02 \x01(\t\x12.\n\x03sdk\x18\x03 \x01(\x0b\x32!.confidence.flags.resolver.v1.Sdk\"\xa4\x07\n\x0c\x46lagAssigned\x12\x12\n\nresolve_id\x18\n \x01(\t\x12=\n\x0b\x63lient_info\x18\x03 \x01(\x0b\x32(.confidence.flags.resolver.v1.ClientInfo\x12\x45\n\x05\x66lags\x18\x0f \x03(\x0b\x32\x36.confidence.flags.resolver.v1.FlagAssigned.AppliedFlag\x1a\xbd\x03\n\x0b\x41ppliedFlag\x12\x0c\n\x04\x66lag\x18\x01 \x01(\t\x12\x15\n\rtargeting_key\x18\x02 \x01(\t\x12\x1e\n\x16targeting_key_selector\x18\x03 \x01(\t\x12T\n\x0f\x61ssignment_info\x18\x04 \x01(\x0b\x32\x39.confidence.flags.resolver.v1.FlagAssigned.AssignmentInfoH\x00\x12Z\n\x12\x64\x65\x66\x61ult_assignment\x18\x05 \x01(\x0b\x32<.confidence.flags.resolver.v1.FlagAssigned.DefaultAssignmentH\x00\x12\x15\n\rassignment_id\x18\x06 \x01(\t\x12\x0c\n\x04rule\x18\x07 \x01(\t\x12T\n\x17\x66\x61llthrough_assignments\x18\x08 \x03(\x0b\x32\x33.confidence.flags.resolver.v1.FallthroughAssignment\x12.\n\napply_time\x18\t \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x0c\n\nassignment\x1a\x32\n\x0e\x41ssignmentInfo\x12\x0f\n\x07segment\x18\x01 \x01(\t\x12\x0f\n\x07variant\x18\x02 \x01(\t\x1a\x85\x02\n\x11\x44\x65\x66\x61ultAssignment\x12\x64\n\x06reason\x18\x01 \x01(\x0e\x32T.confidence.flags.resolver.v1.FlagAssigned.DefaultAssignment.DefaultAssignmentReason\"\x89\x01\n\x17\x44\x65\x66\x61ultAssignmentReason\x12)\n%DEFAULT_ASSIGNMENT_REASON_UNSPECIFIED\x10\x00\x12\x14\n\x10NO_SEGMENT_MATCH\x10\x01\x12\x1a\n\x12NO_TREATMENT_MATCH\x10\x02\x1a\x02\x08\x01\x12\x11\n\rFLAG_ARCHIVED\x10\x03\"s\n\x15\x46\x61llthroughAssignment\x12\x0c\n\x04rule\x18\x01 \x01(\t\x12\x15\n\rassignment_id\x18\x02 \x01(\t\x12\x15\n\rtargeting_key\x18\x03 \x01(\t\x12\x1e\n\x16targeting_key_selector\x18\x04 \x01(\t\"\xdf\x02\n\x11\x43lientResolveInfo\x12\x0e\n\x06\x63lient\x18\x01 \x01(\t\x12\x19\n\x11\x63lient_credential\x18\x02 \x01(\t\x12_\n\x06schema\x18\x03 \x03(\x0b\x32O.confidence.flags.resolver.v1.ClientResolveInfo.EvaluationContextSchemaInstance\x1a\xbd\x01\n\x1f\x45valuationContextSchemaInstance\x12k\n\x06schema\x18\x01 \x03(\x0b\x32[.confidence.flags.resolver.v1.ClientResolveInfo.EvaluationContextSchemaInstance.SchemaEntry\x1a-\n\x0bSchemaEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\"\xb5\x01\n\x0f\x46lagResolveInfo\x12\x0c\n\x04\x66lag\x18\x01 \x01(\t\x12^\n\x14variant_resolve_info\x18\x02 \x03(\x0b\x32@.confidence.flags.resolver.v1.FlagResolveInfo.VariantResolveInfo\x1a\x34\n\x12VariantResolveInfo\x12\x0f\n\x07variant\x18\x01 \x01(\t\x12\r\n\x05\x63ount\x18\x03 \x01(\x03\"]\n\x16WriteOperationsRequest\x12\x43\n\x10store_variant_op\x18\x01 \x03(\x0b\x32).confidence.flags.resolver.v1.VariantData\"\x17\n\x15WriteOperationsResult\"D\n\rVariantReadOp\x12\x0c\n\x04unit\x18\x01 \x01(\t\x12\x17\n\x0fmaterialization\x18\x02 \x01(\t\x12\x0c\n\x04rule\x18\x03 \x01(\t\"8\n\x0fInclusionReadOp\x12\x0c\n\x04unit\x18\x01 \x01(\t\x12\x17\n\x0fmaterialization\x18\x02 \x01(\t\"\xa2\x01\n\x06ReadOp\x12\x46\n\x0fvariant_read_op\x18\x01 \x01(\x0b\x32+.confidence.flags.resolver.v1.VariantReadOpH\x00\x12J\n\x11inclusion_read_op\x18\x02 \x01(\x0b\x32-.confidence.flags.resolver.v1.InclusionReadOpH\x00\x42\x04\n\x02op\"J\n\x15ReadOperationsRequest\x12\x31\n\x03ops\x18\x03 \x03(\x0b\x32$.confidence.flags.resolver.v1.ReadOp\"S\n\x0bVariantData\x12\x0c\n\x04unit\x18\x01 \x01(\t\x12\x17\n\x0fmaterialization\x18\x02 \x01(\t\x12\x0c\n\x04rule\x18\x03 \x01(\t\x12\x0f\n\x07variant\x18\x04 \x01(\t\"K\n\rInclusionData\x12\x0c\n\x04unit\x18\x01 \x01(\t\x12\x17\n\x0fmaterialization\x18\x02 \x01(\t\x12\x13\n\x0bis_included\x18\x03 \x01(\x08\"\xa4\x01\n\nReadResult\x12\x43\n\x0evariant_result\x18\x01 \x01(\x0b\x32).confidence.flags.resolver.v1.VariantDataH\x00\x12G\n\x10inclusion_result\x18\x02 \x01(\x0b\x32+.confidence.flags.resolver.v1.InclusionDataH\x00\x42\x08\n\x06result\"Q\n\x14ReadOperationsResult\x12\x39\n\x07results\x18\x01 \x03(\x0b\x32(.confidence.flags.resolver.v1.ReadResult2\xb2\x03\n\x19InternalFlagLoggerService\x12~\n\x13\x43lientWriteFlagLogs\x12\x32.confidence.flags.resolver.v1.WriteFlagLogsRequest\x1a\x33.confidence.flags.resolver.v1.WriteFlagLogsResponse\x12\x8a\x01\n\x1bWriteMaterializedOperations\x12\x34.confidence.flags.resolver.v1.WriteOperationsRequest\x1a\x33.confidence.flags.resolver.v1.WriteOperationsResult\"\x00\x12\x87\x01\n\x1aReadMaterializedOperations\x12\x33.confidence.flags.resolver.v1.ReadOperationsRequest\x1a\x32.confidence.flags.resolver.v1.ReadOperationsResult\"\x00\x42\xad\x01\n,com.spotify.confidence.sdk.flags.resolver.v1B\x10InternalApiProtoP\x01Zigithub.com/spotify/confidence-resolver/openfeature-provider/go/confidence/internal/proto/resolverinternalb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'confidence.flags.resolver.v1.internal_api_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n,com.spotify.confidence.sdk.flags.resolver.v1B\020InternalApiProtoP\001Zigithub.com/spotify/confidence-resolver/openfeature-provider/go/confidence/internal/proto/resolverinternal'
  _globals['_FLAGASSIGNED_DEFAULTASSIGNMENT_DEFAULTASSIGNMENTREASON'].values_by_name["NO_TREATMENT_MATCH"]._loaded_options = None
  _globals['_FLAGASSIGNED_DEFAULTASSIGNMENT_DEFAULTASSIGNMENTREASON'].values_by_name["NO_TREATMENT_MATCH"]._serialized_options = b'\010\001'
  _globals['_CLIENTRESOLVEINFO_EVALUATIONCONTEXTSCHEMAINSTANCE_SCHEMAENTRY']._loaded_options = None
  _globals['_CLIENTRESOLVEINFO_EVALUATIONCONTEXTSCHEMAINSTANCE_SCHEMAENTRY']._serialized_options = b'8\001'
  _globals['_WRITEFLAGLOGSREQUEST']._serialized_start=157
  _globals['_WRITEFLAGLOGSREQUEST']._serialized_end=467
  _globals['_WRITEFLAGLOGSRESPONSE']._serialized_start=469
  _globals['_WRITEFLAGLOGSRESPONSE']._serialized_end=492
  _globals['_TELEMETRYDATA']._serialized_start=494
  _globals['_TELEMETRYDATA']._serialized_end=557
  _globals['_CLIENTINFO']._serialized_start=559
  _globals['_CLIENTINFO']._serialized_end=662
  _globals['_FLAGASSIGNED']._serialized_start=665
  _globals['_FLAGASSIGNED']._serialized_end=1597
  _globals['_FLAGASSIGNED_APPLIEDFLAG']._serialized_start=836
  _globals['_FLAGASSIGNED_APPLIEDFLAG']._serialized_end=1281
  _globals['_FLAGASSIGNED_ASSIGNMENTINFO']._serialized_start=1283
  _globals['_FLAGASSIGNED_ASSIGNMENTINFO']._serialized_end=1333
  _globals['_FLAGASSIGNED_DEFAULTASSIGNMENT']._serialized_start=1336
  _globals['_FLAGASSIGNED_DEFAULTASSIGNMENT']._serialized_end=1597
  _globals['_FLAGASSIGNED_DEFAULTASSIGNMENT_DEFAULTASSIGNMENTREASON']._serialized_start=1460
  _globals['_FLAGASSIGNED_DEFAULTASSIGNMENT_DEFAULTASSIGNMENTREASON']._serialized_end=1597
  _globals['_FALLTHROUGHASSIGNMENT']._serialized_start=1599
  _globals['_FALLTHROUGHASSIGNMENT']._serialized_end=1714
  _globals['_CLIENTRESOLVEINFO']._serialized_start=1717
  _globals['_CLIENTRESOLVEINFO']._serialized_end=2068
  _globals['_CLIENTRESOLVEINFO_EVALUATIONCONTEXTSCHEMAINSTANCE']._serialized_start=1879
  _globals['_CLIENTRESOLVEINFO_EVALUATIONCONTEXTSCHEMAINSTANCE']._serialized_end=2068
  _globals['_CLIENTRESOLVEINFO_EVALUATIONCONTEXTSCHEMAINSTANCE_SCHEMAENTRY']._serialized_start=2023
  _globals['_CLIENTRESOLVEINFO_EVALUATIONCONTEXTSCHEMAINSTANCE_SCHEMAENTRY']._serialized_end=2068
  _globals['_FLAGRESOLVEINFO']._serialized_start=2071
  _globals['_FLAGRESOLVEINFO']._serialized_end=2252
  _globals['_FLAGRESOLVEINFO_VARIANTRESOLVEINFO']._serialized_start=2200
  _globals['_FLAGRESOLVEINFO_VARIANTRESOLVEINFO']._serialized_end=2252
  _globals['_WRITEOPERATIONSREQUEST']._serialized_start=2254
  _globals['_WRITEOPERATIONSREQUEST']._serialized_end=2347
  _globals['_WRITEOPERATIONSRESULT']._serialized_start=2349
  _globals['_WRITEOPERATIONSRESULT']._serialized_end=2372
  _globals['_VARIANTREADOP']._serialized_start=2374
  _globals['_VARIANTREADOP']._serialized_end=2442
  _globals['_INCLUSIONREADOP']._serialized_start=2444
  _globals['_INCLUSIONREADOP']._serialized_end=2500
  _globals['_READOP']._serialized_start=2503
  _globals['_READOP']._serialized_end=2665
  _globals['_READOPERATIONSREQUEST']._serialized_start=2667
  _globals['_READOPERATIONSREQUEST']._serialized_end=2741
  _globals['_VARIANTDATA']._serialized_start=2743
  _globals['_VARIANTDATA']._serialized_end=2826
  _globals['_INCLUSIONDATA']._serialized_start=2828
  _globals['_INCLUSIONDATA']._serialized_end=2903
  _globals['_READRESULT']._serialized_start=2906
  _globals['_READRESULT']._serialized_end=3070
  _globals['_READOPERATIONSRESULT']._serialized_start=3072
  _globals['_READOPERATIONSRESULT']._serialized_end=3153
  _globals['_INTERNALFLAGLOGGERSERVICE']._serialized_start=3156
  _globals['_INTERNALFLAGLOGGERSERVICE']._serialized_end=3590
# @@protoc_insertion_point(module_scope)



"""Flag logger implementations for Confidence OpenFeature provider.

This module provides flag logging functionality to send flag assignment events
to the Confidence backend via gRPC.
"""

import logging
from concurrent.futures import ThreadPoolExecutor
from typing import Optional, Protocol, runtime_checkable

import grpc

from confidence.proto.confidence.flags.resolver.v1 import (
    internal_api_pb2,
    internal_api_pb2_grpc,
)

logger = logging.getLogger(__name__)

# gRPC target for the Confidence edge service
GRPC_TARGET = "edge-grpc.spotify.com:443"


@runtime_checkable
class FlagLogger(Protocol):
    """Protocol for flag logging."""

    def write(self, request_bytes: bytes) -> None:
        """Write flag logs asynchronously.

        Args:
            request_bytes: Serialized WriteFlagLogsRequest proto bytes.
        """
        ...

    def shutdown(self) -> None:
        """Shutdown the logger and wait for pending writes to complete."""
        ...


class GrpcFlagLogger:
    """gRPC-based flag logger that sends flag logs to the Confidence backend.

    Writes are performed asynchronously using a thread pool. The logger
    skips empty requests (no flag_assigned, client_resolve_info, or flag_resolve_info).
    """

    def __init__(
        self,
        client_secret: str,
        channel: Optional[grpc.Channel] = None,
    ) -> None:
        """Initialize the gRPC flag logger.

        Args:
            client_secret: The Confidence client secret for authentication.
            channel: Optional gRPC channel for testing. If not provided,
                    a secure channel to GRPC_TARGET will be created.
        """
        self._client_secret = client_secret
        self._executor = ThreadPoolExecutor(max_workers=2)

        if channel is not None:
            self._channel = channel
            self._owns_channel = False
        else:
            self._channel = grpc.secure_channel(
                GRPC_TARGET,
                grpc.ssl_channel_credentials(),
            )
            self._owns_channel = True

        self._stub = internal_api_pb2_grpc.InternalFlagLoggerServiceStub(self._channel)

    def write(self, request_bytes: bytes) -> None:
        """Write flag logs asynchronously.

        Skips empty requests (no data).

        Args:
            request_bytes: Serialized WriteFlagLogsRequest proto bytes.
        """
        # Skip empty bytes
        if not request_bytes:
            logger.debug("Skipping empty flag log request (empty bytes)")
            return

        # Parse the request to check if it has any data
        try:
            request = internal_api_pb2.WriteFlagLogsRequest()
            request.ParseFromString(request_bytes)
        except Exception as e:
            logger.error("Failed to parse WriteFlagLogsRequest: %s", e)
            return

        # Skip if all lists are empty
        if (
            len(request.flag_assigned) == 0
            and len(request.client_resolve_info) == 0
            and len(request.flag_resolve_info) == 0
        ):
            logger.debug("Skipping empty flag log request (no data)")
            return

        # Submit async write
        self._executor.submit(self._send_request, request)

    def _send_request(self, request: internal_api_pb2.WriteFlagLogsRequest) -> None:
        """Send the request to the backend (runs in thread pool).

        Args:
            request: The WriteFlagLogsRequest to send.
        """
        try:
            metadata = [("authorization", f"ClientSecret {self._client_secret}")]
            self._stub.ClientWriteFlagLogs(request, metadata=metadata, timeout=30.0)
            logger.debug(
                "Successfully sent flag log with %d entries",
                len(request.flag_assigned),
            )
        except Exception as e:
            logger.error("Failed to write flag logs: %s", e)

    def shutdown(self) -> None:
        """Shutdown the logger and wait for pending writes to complete."""
        self._executor.shutdown(wait=True)
        if self._owns_channel:
            self._channel.close()


class NoOpFlagLogger:
    """A no-op flag logger that drops all requests.

    Useful for testing or when flag logging should be disabled.
    """

    def write(self, request_bytes: bytes) -> None:
        """Drop the request (do nothing).

        Args:
            request_bytes: Ignored.
        """
        pass

    def shutdown(self) -> None:
        """Do nothing (no resources to clean up)."""
        pass



"""Local resolver with crash recovery.

This module provides the LocalResolver class that wraps WasmResolver
with automatic crash recovery and log buffering.
"""

import logging
from typing import List, Optional, Tuple

from wasmtime import Trap as WasmTrap

from confidence.proto.confidence.wasm import wasm_api_pb2
from confidence.wasm_resolver import WasmResolver

logger = logging.getLogger(__name__)

# Exception types that indicate a WASM crash requiring reload
WasmCrashError = (RuntimeError, WasmTrap)


class LocalResolver:
    """Local flag resolver with crash recovery.

    Wraps WasmResolver and provides:
    - Automatic WASM reload on RuntimeError
    - State caching for recovery
    - Log buffering on crash before reload

    This follows the same pattern as the JavaScript WasmResolver class
    and Go RecoveringResolver.
    """

    def __init__(
        self,
        wasm_bytes: bytes,
        delegate_factory: Optional[type] = None,
    ) -> None:
        """Initialize the local resolver.

        Args:
            wasm_bytes: The compiled WASM binary bytes.
            delegate_factory: Optional factory for creating delegate resolvers.
                Defaults to WasmResolver.
        """
        self._wasm_bytes = wasm_bytes
        self._delegate_factory = delegate_factory or WasmResolver
        self._delegate: WasmResolver = self._delegate_factory(wasm_bytes)
        self._current_state: Optional[Tuple[bytes, str, object]] = None
        self._buffered_logs: List[bytes] = []

    def _reload_instance(self, error: BaseException) -> None:
        """Reload the WASM instance after a failure.

        Attempts to flush logs from the failing instance before reloading.

        Args:
            error: The error that caused the reload.
        """
        logger.error("Failure calling into wasm: %s", error)

        # Try to flush logs before discarding the old instance
        try:
            logs = self._delegate.flush_logs()
            if logs:
                self._buffered_logs.append(logs)
        except Exception:
            logger.error("Failed to flush_logs on error")

        # Create new delegate
        self._delegate = self._delegate_factory(self._wasm_bytes)

        # Restore state if available
        if self._current_state is not None:
            state, account_id, sdk = self._current_state
            try:
                self._delegate.set_resolver_state(state, account_id, sdk)
            except Exception as e:
                logger.error("Failed to restore state after reload: %s", e)

    def set_resolver_state(
        self, state: bytes, account_id: str, sdk: Optional[object] = None
    ) -> None:
        """Set the resolver state.

        Stores the state for recovery and delegates to WasmResolver.

        Args:
            state: The serialized resolver state bytes.
            account_id: The account ID for the resolver.
            sdk: Optional SDK identifier and version.
        """
        self._current_state = (state, account_id, sdk)
        try:
            self._delegate.set_resolver_state(state, account_id, sdk)
        except WasmCrashError as error:
            self._reload_instance(error)
            raise

    def resolve_process(
        self, request: wasm_api_pb2.ResolveProcessRequest
    ) -> wasm_api_pb2.ResolveProcessResponse:
        """Resolve flags using the WASM module.

        On WASM crash (RuntimeError or Trap), reloads the WASM instance and re-raises.

        Args:
            request: The resolve process request.

        Returns:
            The resolve process response.

        Raises:
            RuntimeError: If the WASM module encounters an error.
            WasmTrap: If the WASM module hits a trap (e.g., unreachable).
        """
        try:
            return self._delegate.resolve_process(request)
        except WasmCrashError as error:
            self._reload_instance(error)
            raise

    def register_resolve(self, request: wasm_api_pb2.RegisterResolveRequest) -> None:
        """Register a resolve evaluation for telemetry."""
        try:
            self._delegate.register_resolve(request)
        except WasmCrashError:
            pass

    def flush_logs(self) -> bytes:
        """Flush all pending logs including buffered logs.

        Returns:
            The combined serialized log bytes.
        """
        try:
            current_logs = self._delegate.flush_logs()
            self._buffered_logs.append(current_logs)

            # Combine all buffered logs
            total_length = sum(len(chunk) for chunk in self._buffered_logs)
            result = bytearray(total_length)
            offset = 0
            for chunk in self._buffered_logs:
                result[offset : offset + len(chunk)] = chunk
                offset += len(chunk)

            # Clear buffer
            self._buffered_logs.clear()

            return bytes(result)
        except WasmCrashError as error:
            self._reload_instance(error)
            raise

    def flush_assigned(self) -> bytes:
        """Flush pending assignment logs.

        Returns:
            The serialized assignment data bytes.
        """
        # TODO: Buffer logs and resend on failure (like JS implementation)
        return self._delegate.flush_assigned()

    def prometheus_snapshot(self) -> str:
        """Get a Prometheus metrics snapshot.

        Returns:
            The Prometheus metrics text, or empty string on failure.
        """
        try:
            return self._delegate.prometheus_snapshot()
        except WasmCrashError as error:
            logger.warning("prometheus snapshot failed: %s", error)
            self._reload_instance(error)
            return ""



"""Materialization store module for Confidence OpenFeature provider.

This module provides storage abstractions for materialization data used in
flag resolution.

Materializations support two key use cases:
1. Sticky Assignments: Maintain consistent variant assignments across evaluations
   even when targeting attributes change.
2. Custom Targeting via Materialized Segments: Precomputed sets of identifiers from
   datasets that should be targeted.
"""

import logging
from dataclasses import dataclass
from typing import List, Optional, Protocol, Union

import grpc

from confidence.proto.confidence.flags.resolver.v1 import (
    internal_api_pb2,
    internal_api_pb2_grpc,
)

logger = logging.getLogger(__name__)


@dataclass
class VariantReadOp:
    """Query operation to retrieve the variant assignment for a unit and rule.

    Used for sticky assignments to fetch the previously assigned variant.

    Attributes:
        unit: The entity identifier (user ID, session ID, etc.)
        materialization: The materialization context identifier
        rule: The targeting rule identifier
    """

    unit: str
    materialization: str
    rule: str


@dataclass
class VariantReadResult:
    """Result containing the variant assignment for a unit and rule.

    Used for sticky assignments - returns the previously assigned variant for a
    unit and targeting rule combination.

    Attributes:
        unit: The entity identifier (user ID, session ID, etc.)
        materialization: The materialization context identifier
        rule: The targeting rule identifier
        variant: The assigned variant name, or None if no assignment exists
    """

    unit: str
    materialization: str
    rule: str
    variant: Optional[str] = None


@dataclass
class VariantWriteOp:
    """A variant assignment write operation.

    Used to store sticky variant assignments, recording which variant a unit
    (user, session, etc.) should receive for a specific targeting rule.

    Attributes:
        unit: The entity identifier (user ID, session ID, etc.)
        materialization: The materialization context identifier
        rule: The targeting rule identifier
        variant: The assigned variant name
    """

    unit: str
    materialization: str
    rule: str
    variant: str


@dataclass
class InclusionReadOp:
    """Query operation to check if a unit is included in a materialized segment.

    Used for custom targeting via materialized segments - checks if a unit
    is part of a precomputed set of identifiers.

    Attributes:
        unit: The entity identifier (user ID, session ID, etc.)
        materialization: The materialization context identifier
    """

    unit: str
    materialization: str


@dataclass
class InclusionReadResult:
    """Result containing whether a unit is included in a materialized segment.

    Attributes:
        unit: The entity identifier (user ID, session ID, etc.)
        materialization: The materialization context identifier
        included: Whether the unit is included in the segment
    """

    unit: str
    materialization: str
    included: bool


class MaterializationNotSupportedError(Exception):
    """Raised when a MaterializationStore doesn't support the requested operation.

    This triggers the provider to fall back to remote gRPC resolution via the
    Confidence service, which manages materializations server-side.
    """

    def __init__(
        self, message: str = "materialization operation not supported"
    ) -> None:
        super().__init__(message)
        self.message = message


# Type aliases for read operations and results
ReadOp = Union[VariantReadOp, InclusionReadOp]
ReadResult = Union[VariantReadResult, InclusionReadResult]


class MaterializationStore(Protocol):
    """Storage abstraction for materialization data used in flag resolution.

    Materializations support two key use cases:
    1. Sticky Assignments: Maintain consistent variant assignments across
       evaluations even when targeting attributes change.
    2. Custom Targeting via Materialized Segments: Precomputed sets of
       identifiers that should be targeted.

    Thread Safety: Implementations must be thread-safe as they may be called
    concurrently from multiple threads resolving flags in parallel.
    """

    def read(self, ops: List[ReadOp]) -> List[ReadResult]:
        """Perform a batch read of materialization data.

        Args:
            ops: The list of read operations to perform
                (VariantReadOp or InclusionReadOp)

        Returns:
            The read results (VariantReadResult or InclusionReadResult)

        Raises:
            MaterializationNotSupportedError: If the store doesn't support reads
        """
        ...

    def write(self, ops: List[VariantWriteOp]) -> None:
        """Perform a batch write of materialization data.

        Args:
            ops: The list of write operations to perform

        Raises:
            MaterializationNotSupportedError: If the store doesn't support writes
        """
        ...


class UnsupportedMaterializationStore:
    """A MaterializationStore that always raises MaterializationNotSupportedError.

    This is the default store used by the provider to trigger fallback to remote
    gRPC resolution when materializations are needed.
    """

    def read(self, ops: List[ReadOp]) -> List[ReadResult]:
        """Read always raises MaterializationNotSupportedError.

        Args:
            ops: The list of read operations to perform

        Raises:
            MaterializationNotSupportedError: Always raised
        """
        raise MaterializationNotSupportedError("materialization read not supported")

    def write(self, ops: List[VariantWriteOp]) -> None:
        """Write always raises MaterializationNotSupportedError.

        Args:
            ops: The list of write operations to perform

        Raises:
            MaterializationNotSupportedError: Always raised
        """
        raise MaterializationNotSupportedError("materialization write not supported")


class RemoteMaterializationStore:
    """A MaterializationStore that stores data remotely via gRPC.

    This implementation is useful when you want the Confidence service to manage
    materialization storage server-side rather than maintaining local state.
    """

    GRPC_TARGET = "edge-grpc.spotify.com:443"

    def __init__(
        self,
        client_secret: str,
        channel: Optional[grpc.Channel] = None,
    ) -> None:
        """Initialize the remote materialization store.

        Args:
            client_secret: The client secret for authentication
            channel: Optional gRPC channel. If not provided, a secure channel
                    will be created to GRPC_TARGET.
        """
        self._client_secret = client_secret
        if channel is None:
            channel = grpc.secure_channel(
                self.GRPC_TARGET,
                grpc.ssl_channel_credentials(),
            )
            self._owns_channel = True
        else:
            self._owns_channel = False
        self._channel = channel
        self._stub = internal_api_pb2_grpc.InternalFlagLoggerServiceStub(channel)  # type: ignore[no-untyped-call]

    def _get_metadata(self) -> List[tuple[str, str]]:
        """Get the authorization metadata for gRPC calls."""
        return [("authorization", f"ClientSecret {self._client_secret}")]

    def read(self, ops: List[ReadOp]) -> List[ReadResult]:
        """Perform a batch read of materialization data from the remote service.

        Args:
            ops: The list of read operations to perform
                (VariantReadOp or InclusionReadOp)

        Returns:
            The read results (VariantReadResult or InclusionReadResult)
        """
        if not ops:
            return []

        # Convert to proto format
        proto_ops = []
        for op in ops:
            proto_op = internal_api_pb2.ReadOp()  # type: ignore[attr-defined]
            if isinstance(op, VariantReadOp):
                proto_op.variant_read_op.CopyFrom(
                    internal_api_pb2.VariantReadOp(  # type: ignore[attr-defined]
                        unit=op.unit,
                        materialization=op.materialization,
                        rule=op.rule,
                    )
                )
            elif isinstance(op, InclusionReadOp):
                proto_op.inclusion_read_op.CopyFrom(
                    internal_api_pb2.InclusionReadOp(  # type: ignore[attr-defined]
                        unit=op.unit,
                        materialization=op.materialization,
                    )
                )
            proto_ops.append(proto_op)

        # Build request
        request = internal_api_pb2.ReadOperationsRequest(  # type: ignore[attr-defined]
            ops=proto_ops
        )

        # Call gRPC service
        try:
            response = self._stub.ReadMaterializedOperations(
                request,
                metadata=self._get_metadata(),
                timeout=30.0,
            )
        except grpc.RpcError as e:
            logger.error("Failed to read materialized operations: %s", e)
            raise

        logger.debug("Read %d materialization operations", len(ops))

        # Convert results
        results: List[ReadResult] = []
        for proto_result in response.results:
            if proto_result.HasField("variant_result"):
                variant_data = proto_result.variant_result
                results.append(
                    VariantReadResult(
                        unit=variant_data.unit,
                        materialization=variant_data.materialization,
                        rule=variant_data.rule,
                        variant=variant_data.variant if variant_data.variant else None,
                    )
                )
            elif proto_result.HasField("inclusion_result"):
                inclusion_data = proto_result.inclusion_result
                results.append(
                    InclusionReadResult(
                        unit=inclusion_data.unit,
                        materialization=inclusion_data.materialization,
                        included=inclusion_data.is_included,
                    )
                )
        return results

    def write(self, ops: List[VariantWriteOp]) -> None:
        """Perform a batch write of materialization data to the remote service.

        Args:
            ops: The list of write operations to perform
        """
        if not ops:
            return

        # Convert to proto format
        proto_ops = []
        for op in ops:
            proto_ops.append(
                internal_api_pb2.VariantData(  # type: ignore[attr-defined]
                    unit=op.unit,
                    materialization=op.materialization,
                    rule=op.rule,
                    variant=op.variant,
                )
            )

        # Build request
        request = internal_api_pb2.WriteOperationsRequest(  # type: ignore[attr-defined]
            store_variant_op=proto_ops
        )

        # Call gRPC service
        try:
            self._stub.WriteMaterializedOperations(
                request,
                metadata=self._get_metadata(),
                timeout=30.0,
            )
        except grpc.RpcError as e:
            logger.error("Failed to write materialized operations: %s", e)
            raise

        logger.debug("Wrote %d materialization operations", len(ops))

    def close(self) -> None:
        """Close the underlying gRPC channel if owned."""
        if self._owns_channel:
            self._channel.close()



"""Confidence OpenFeature provider implementation.

This module provides the ConfidenceProvider class that implements the OpenFeature
AbstractProvider interface for local flag resolution using the Confidence WASM resolver.
"""

import logging
import threading
import time
from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar

import grpc
import httpx
from google.protobuf import struct_pb2
from openfeature.evaluation_context import EvaluationContext
from openfeature.event import ProviderEventDetails
from openfeature.exception import ErrorCode
from openfeature.flag_evaluation import FlagResolutionDetails, Reason
from openfeature.provider import AbstractProvider, Metadata, ProviderStatus

from confidence.flag_logger import FlagLogger
from confidence.local_resolver import LocalResolver
from confidence.materialization import (
    InclusionReadOp,
    InclusionReadResult,
    MaterializationNotSupportedError,
    MaterializationStore,
    ReadOp,
    RemoteMaterializationStore,
    UnsupportedMaterializationStore,
    VariantReadOp,
    VariantReadResult,
    VariantWriteOp,
)
from confidence.proto.confidence.flags.resolver.v1 import api_pb2, types_pb2
from confidence.proto.confidence.wasm import wasm_api_pb2
from confidence.state_fetcher import StateFetcher
from confidence.version import __version__

# Type variable for generic resolution
T = TypeVar("T")

logger = logging.getLogger(__name__)

# Default intervals
DEFAULT_STATE_POLL_INTERVAL = 30.0
DEFAULT_LOG_POLL_INTERVAL = 15.0
DEFAULT_ASSIGN_POLL_INTERVAL = 0.1


class SnapshotConfig:
    """Configuration for :meth:`ConfidenceProvider.get_prometheus_metrics`.

    **Experimental:** this API is subject to change.
    """

    pass


def _load_wasm_from_resources() -> bytes:
    """Load the WASM binary from package resources.

    Returns:
        The WASM binary bytes.

    Raises:
        FileNotFoundError: If the WASM binary cannot be found.
    """
    try:
        # Try importlib.resources first (Python 3.9+)
        import importlib.resources as resources

        try:
            # Python 3.9+ with files()
            files = resources.files("confidence")
            wasm_path = files.joinpath("wasm").joinpath("confidence_resolver.wasm")
            return wasm_path.read_bytes()
        except (AttributeError, FileNotFoundError, TypeError):
            pass

    except ImportError:
        pass

    # Fallback to pkg_resources
    try:
        import pkg_resources

        return pkg_resources.resource_string(
            "confidence", "wasm/confidence_resolver.wasm"
        )
    except Exception:
        pass

    # Development fallback: try resources directory
    from pathlib import Path

    dev_path = (
        Path(__file__).parent.parent.parent
        / "resources"
        / "wasm"
        / "confidence_resolver.wasm"
    )
    if dev_path.exists():
        return dev_path.read_bytes()

    raise FileNotFoundError(
        "Could not find confidence_resolver.wasm in package resources"
    )


class ConfidenceProvider(AbstractProvider):
    """Confidence OpenFeature provider for local flag resolution.

    This provider uses a WASM-based resolver for local flag evaluation,
    with background threads for state polling and log flushing.

    Attributes:
        PROVIDER_NAME: The name of this provider.
    """

    PROVIDER_NAME = "confidence-sdk-python-local"

    def __init__(
        self,
        client_secret: str,
        state_poll_interval: float = DEFAULT_STATE_POLL_INTERVAL,
        log_poll_interval: float = DEFAULT_LOG_POLL_INTERVAL,
        assign_poll_interval: float = DEFAULT_ASSIGN_POLL_INTERVAL,
        materialization_store: Optional[MaterializationStore] = None,
        use_remote_materialization_store: bool = False,
        http_client: Optional[httpx.Client] = None,
        grpc_channel: Optional[grpc.Channel] = None,
        state_fetcher: Optional[StateFetcher] = None,
        flag_logger: Optional[FlagLogger] = None,
        wasm_bytes: Optional[bytes] = None,
    ) -> None:
        """Initialize the Confidence provider.

        Args:
            client_secret: The Confidence client secret (required).
            state_poll_interval: Interval for state polling (default: 30.0).
            log_poll_interval: Interval for log flushing (default: 10.0).
            assign_poll_interval: Interval for assignment flushing (default: 0.1).
            materialization_store: Optional custom materialization store.
            use_remote_materialization_store: Use remote store for materializations.
            http_client: Optional httpx.Client for StateFetcher.
            grpc_channel: Optional grpc.Channel for FlagLogger and MaterializationStore.
            state_fetcher: Optional state fetcher for testing.
            flag_logger: Optional flag logger for testing.
            wasm_bytes: Optional WASM bytes for testing.
        """
        self._client_secret = client_secret
        self._state_poll_interval = state_poll_interval
        self._log_poll_interval = log_poll_interval
        self._assign_poll_interval = assign_poll_interval
        self._http_client = http_client
        self._grpc_channel = grpc_channel

        # Initialize resolver (created during initialize())
        self._resolver: Optional[LocalResolver] = None
        self._resolver_lock = threading.Lock()

        # WASM bytes (loaded lazily or from test)
        self._wasm_bytes = wasm_bytes

        # State fetcher (injected or created)
        self._state_fetcher = state_fetcher

        # Flag logger (injected or created)
        self._flag_logger = flag_logger

        # Materialization store
        if materialization_store is not None:
            self._materialization_store: MaterializationStore = materialization_store
        elif use_remote_materialization_store:
            self._materialization_store = RemoteMaterializationStore(
                client_secret=client_secret,
                channel=grpc_channel,
            )
        else:
            self._materialization_store = UnsupportedMaterializationStore()

        # Background threads
        self._shutdown_event = threading.Event()
        self._state_thread: Optional[threading.Thread] = None
        self._log_thread: Optional[threading.Thread] = None

        # Provider status
        self._status = ProviderStatus.NOT_READY

    def get_status(self) -> ProviderStatus:
        """Get the current provider status.

        Returns:
            The current provider status (NOT_READY, READY, ERROR, STALE, FATAL).
        """
        return self._status

    def get_metadata(self) -> Metadata:
        """Get provider metadata.

        Returns:
            Metadata with the provider name.
        """
        return Metadata(name=self.PROVIDER_NAME)

    def initialize(self, evaluation_context: EvaluationContext) -> None:
        """Initialize the provider.

        Loads the WASM module, fetches initial state, and starts background threads.

        Args:
            evaluation_context: The initial evaluation context.

        Raises:
            Exception: If initialization fails.
        """
        # Load WASM bytes if not provided
        if self._wasm_bytes is None:
            self._wasm_bytes = _load_wasm_from_resources()

        # Create resolver
        self._resolver = LocalResolver(self._wasm_bytes)

        # Create state fetcher if not injected
        if self._state_fetcher is None:
            from confidence.state_fetcher import StateFetcher

            self._state_fetcher = StateFetcher(
                client_secret=self._client_secret,
                http_client=self._http_client,
            )

        # Create flag logger if not injected
        if self._flag_logger is None:
            from confidence.flag_logger import GrpcFlagLogger

            self._flag_logger = GrpcFlagLogger(
                client_secret=self._client_secret,
                channel=self._grpc_channel,
            )

        # Fetch initial state - don't fail if this fails, background thread will retry
        try:
            state, account_id, _ = self._state_fetcher.fetch()
            if account_id:
                sdk = types_pb2.Sdk(
                    id=types_pb2.SdkId.SDK_ID_PYTHON_PROVIDER,
                    version=__version__,
                )
                self._resolver.set_resolver_state(state, account_id, sdk)
                self._status = ProviderStatus.READY
                self.emit_provider_ready(ProviderEventDetails())
                logger.info("ConfidenceProvider initialized successfully")
            else:
                logger.warning(
                    "Initial state load returned empty account ID, "
                    "provider starting in NOT_READY state"
                )
        except Exception as e:
            logger.warning(
                "Initial state load failed, provider starting in NOT_READY state: %s",
                e,
            )

        # Start background threads (will retry state fetch if needed)
        self._start_background_threads()

    def shutdown(self) -> None:
        """Shutdown the provider.

        Stops background threads, flushes logs, and cleans up resources.
        """
        logger.info("Shutting down ConfidenceProvider")

        # Set status to NOT_READY
        self._status = ProviderStatus.NOT_READY

        # Signal shutdown to background threads
        self._shutdown_event.set()

        # Wait for threads to finish
        if self._state_thread is not None:
            self._state_thread.join(timeout=5.0)
            self._state_thread = None

        if self._log_thread is not None:
            self._log_thread.join(timeout=5.0)
            self._log_thread = None

        # Flush final logs
        if self._resolver is not None:
            try:
                log_data = self._resolver.flush_logs()
                if log_data and self._flag_logger is not None:
                    self._flag_logger.write(log_data)
            except Exception as e:
                logger.error("Failed to flush final logs: %s", e)

        # Shutdown flag logger
        if self._flag_logger is not None:
            self._flag_logger.shutdown()

        # Close materialization store if it exposes a close method
        close_store = getattr(self._materialization_store, "close", None)
        if callable(close_store):
            try:
                close_store()
            except Exception as e:
                logger.error("Failed to close materialization store: %s", e)

        logger.info("ConfidenceProvider shutdown complete")

    def _resolve_typed(
        self,
        flag_key: str,
        default_value: T,
        evaluation_context: Optional[EvaluationContext],
        type_check: Callable[[Any], bool],
        type_convert: Callable[[Any], T],
    ) -> FlagResolutionDetails[T]:
        """Generic typed resolution with type checking and conversion."""
        start_time = time.monotonic()
        result = self._resolve_object(
            flag_key, default_value, evaluation_context, start_time
        )

        if result.value is None or result.error_code is not None:
            return FlagResolutionDetails(
                value=default_value,
                reason=result.reason,
                error_code=result.error_code,
                error_message=result.error_message,
                variant=result.variant,
            )

        if not type_check(result.value):
            self._do_register_resolve(
                types_pb2.RESOLVE_REASON_TYPE_MISMATCH, start_time
            )
            return FlagResolutionDetails(
                value=default_value,
                reason=Reason.ERROR,
                error_code=ErrorCode.TYPE_MISMATCH,
                error_message=f"Value is not {type(default_value).__name__}",
            )

        return FlagResolutionDetails(
            value=type_convert(result.value),
            reason=result.reason,
            variant=result.variant,
        )

    def resolve_boolean_details(
        self,
        flag_key: str,
        default_value: bool,
        evaluation_context: Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[bool]:
        """Resolve a boolean flag."""
        return self._resolve_typed(
            flag_key,
            default_value,
            evaluation_context,
            type_check=lambda v: isinstance(v, bool),
            type_convert=lambda v: v,
        )

    def resolve_string_details(
        self,
        flag_key: str,
        default_value: str,
        evaluation_context: Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[str]:
        """Resolve a string flag."""
        return self._resolve_typed(
            flag_key,
            default_value,
            evaluation_context,
            type_check=lambda v: isinstance(v, str),
            type_convert=lambda v: v,
        )

    def resolve_integer_details(
        self,
        flag_key: str,
        default_value: int,
        evaluation_context: Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[int]:
        """Resolve an integer flag."""
        return self._resolve_typed(
            flag_key,
            default_value,
            evaluation_context,
            # Accept int (but not bool) or float that is a whole number
            type_check=lambda v: (
                (isinstance(v, int) and not isinstance(v, bool))
                or (isinstance(v, float) and v.is_integer())
            ),
            type_convert=lambda v: int(v),
        )

    def resolve_float_details(
        self,
        flag_key: str,
        default_value: float,
        evaluation_context: Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[float]:
        """Resolve a float flag."""
        return self._resolve_typed(
            flag_key,
            default_value,
            evaluation_context,
            # Accept int or float (but not bool)
            type_check=lambda v: (
                isinstance(v, (int, float)) and not isinstance(v, bool)
            ),
            type_convert=lambda v: float(v),
        )

    def resolve_object_details(
        self,
        flag_key: str,
        default_value: Dict[str, Any],
        evaluation_context: Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[Dict[str, Any]]:
        """Resolve an object flag."""
        return self._resolve_typed(
            flag_key,
            default_value,
            evaluation_context,
            type_check=lambda v: isinstance(v, dict),
            type_convert=lambda v: v,
        )

    def _resolve_object(
        self,
        flag_key: str,
        default_value: Any,
        evaluation_context: Optional[EvaluationContext],
        start_time: float,
    ) -> FlagResolutionDetails[Any]:
        """Core resolution logic for all flag types."""

        if self._resolver is None:
            return FlagResolutionDetails(
                value=default_value,
                reason=Reason.ERROR,
                error_code=ErrorCode.PROVIDER_NOT_READY,
                error_message="Provider not initialized",
            )

        try:
            flag_name, path = self._parse_flag_path(flag_key)
            proto_context = self._context_to_proto(evaluation_context)

            resolve_req = api_pb2.ResolveFlagsRequest()
            resolve_req.flags.append(f"flags/{flag_name}")
            resolve_req.client_secret = self._client_secret
            resolve_req.apply = True
            if proto_context is not None:
                resolve_req.evaluation_context.CopyFrom(proto_context)
            resolve_req.sdk.id = types_pb2.SdkId.SDK_ID_PYTHON_PROVIDER
            resolve_req.sdk.version = __version__

            request = wasm_api_pb2.ResolveProcessRequest()
            if self._materialization_store is not None and not isinstance(
                self._materialization_store, UnsupportedMaterializationStore
            ):
                request.deferred_materializations.CopyFrom(resolve_req)
            else:
                request.without_materializations.CopyFrom(resolve_req)

            response = self._resolve_with_materialization(request)

            if not response.HasField("resolved"):
                return FlagResolutionDetails(
                    value=default_value,
                    reason=Reason.ERROR,
                    error_code=ErrorCode.GENERAL,
                    error_message="Unexpected suspended response",
                )

            resolved_flags = response.resolved.response.resolved_flags
            if len(resolved_flags) == 0:
                self._do_register_resolve(
                    types_pb2.RESOLVE_REASON_FLAG_NOT_FOUND, start_time
                )
                return FlagResolutionDetails(
                    value=default_value,
                    reason=Reason.ERROR,
                    error_code=ErrorCode.FLAG_NOT_FOUND,
                    error_message=f"Flag '{flag_name}' not found",
                )

            resolved_flag = resolved_flags[0]

            expected_name = f"flags/{flag_name}"
            if resolved_flag.flag != expected_name:
                self._do_register_resolve(
                    types_pb2.RESOLVE_REASON_FLAG_NOT_FOUND, start_time
                )
                return FlagResolutionDetails(
                    value=default_value,
                    reason=Reason.ERROR,
                    error_code=ErrorCode.FLAG_NOT_FOUND,
                    error_message="Unexpected flag returned",
                )

            if (
                resolved_flag.reason
                == types_pb2.ResolveReason.RESOLVE_REASON_MATERIALIZATION_NOT_SUPPORTED
            ):
                logger.warning(
                    "Flag '%s' requires materializations but no materialization store is "
                    "configured. Enable it via ConfidenceProvider(use_remote_materialization_store=True)",
                    flag_name,
                )
                self._do_register_resolve(resolved_flag.reason, start_time)
                return FlagResolutionDetails(
                    value=default_value,
                    reason=Reason.ERROR,
                    error_code=ErrorCode.GENERAL,
                    error_message=f"Flag '{flag_name}' requires materializations. Configure a materialization store.",
                )

            if not resolved_flag.variant:
                self._do_register_resolve(resolved_flag.reason, start_time)
                return FlagResolutionDetails(
                    value=default_value,
                    reason=self._map_resolve_reason(resolved_flag.reason),
                )

            value = self._proto_struct_to_dict(resolved_flag.value)

            if path:
                value, found = self._get_value_for_path(path, value)
                if not found:
                    self._do_register_resolve(
                        types_pb2.RESOLVE_REASON_FLAG_NOT_FOUND, start_time
                    )
                    return FlagResolutionDetails(
                        value=default_value,
                        reason=Reason.ERROR,
                        error_code=ErrorCode.FLAG_NOT_FOUND,
                        error_message=f"Path '{path}' not found in flag '{flag_name}'",
                    )

            self._do_register_resolve(resolved_flag.reason, start_time)
            return FlagResolutionDetails(
                value=value,
                reason=self._map_resolve_reason(resolved_flag.reason),
                variant=resolved_flag.variant,
            )

        except Exception as e:
            logger.error("Failed to resolve flag '%s': %s", flag_key, e)
            return FlagResolutionDetails(
                value=default_value,
                reason=Reason.ERROR,
                error_code=ErrorCode.GENERAL,
                error_message=str(e),
            )

    def _do_register_resolve(self, reason: int, start_time: float) -> None:
        """Register a resolve evaluation for telemetry."""
        if self._resolver is None:
            return
        latency_us = int((time.monotonic() - start_time) * 1_000_000)
        try:
            request = wasm_api_pb2.RegisterResolveRequest()
            request.reason = reason
            request.latency_us = min(latency_us, 2**32 - 1)
            with self._resolver_lock:
                self._resolver.register_resolve(request)
        except Exception:
            logger.warning("Failed to register resolve telemetry", exc_info=True)

    def _resolve_with_materialization(
        self, request: wasm_api_pb2.ResolveProcessRequest
    ) -> wasm_api_pb2.ResolveProcessResponse:
        """Resolve with materialization handling (suspend/resume).

        Args:
            request: The resolve process request.

        Returns:
            The resolve process response.

        Raises:
            RuntimeError: If resolution fails.
        """
        with self._resolver_lock:
            response = self._resolver.resolve_process(request)

        return self._handle_response(response)

    def _handle_response(
        self, response: wasm_api_pb2.ResolveProcessResponse
    ) -> wasm_api_pb2.ResolveProcessResponse:
        """Handle initial resolve response."""
        if response.HasField("resolved"):
            if response.resolved.materializations_to_write:
                self._write_materializations(
                    response.resolved.materializations_to_write
                )
            return response

        if response.HasField("suspended"):
            # Read materializations from store
            try:
                materializations = self._read_materializations(
                    response.suspended.materializations_to_read
                )
            except MaterializationNotSupportedError as e:
                raise RuntimeError(f"failed to read materializations: {e.message}")

            # Build resume request
            resume_request = wasm_api_pb2.ResolveProcessRequest()
            resume_request.resume.state = response.suspended.state
            for mat in materializations:
                resume_request.resume.materializations.append(mat)

            with self._resolver_lock:
                resume_response = self._resolver.resolve_process(resume_request)

            return self._handle_resume_response(resume_response)

        raise RuntimeError("Unexpected empty resolve response")

    def _handle_resume_response(
        self, response: wasm_api_pb2.ResolveProcessResponse
    ) -> wasm_api_pb2.ResolveProcessResponse:
        """Handle response after resume - should not suspend again."""
        if response.HasField("resolved"):
            if response.resolved.materializations_to_write:
                self._write_materializations(
                    response.resolved.materializations_to_write
                )
            return response

        if response.HasField("suspended"):
            raise RuntimeError("Unexpected second suspend after resume")

        raise RuntimeError("Unexpected empty resolve response after resume")

    def _read_materializations(
        self, records: List[wasm_api_pb2.MaterializationRecord]
    ) -> List[wasm_api_pb2.MaterializationRecord]:
        """Read materializations from the store.

        Converts MaterializationRecords to ReadOps, queries the store,
        then converts results back to MaterializationRecords for a Resume request.

        Args:
            records: The MaterializationRecords from a Suspended response.

        Returns:
            List of MaterializationRecords for the Resume request.
        """
        ops: List[ReadOp] = []
        for record in records:
            if record.rule:
                ops.append(
                    VariantReadOp(
                        unit=record.unit,
                        materialization=record.materialization,
                        rule=record.rule,
                    )
                )
            else:
                ops.append(
                    InclusionReadOp(
                        unit=record.unit,
                        materialization=record.materialization,
                    )
                )

        results = self._materialization_store.read(ops)

        # Convert results back to MaterializationRecords.
        # Records with empty variant and "not included" inclusion results are omitted
        # (absence = no prior assignment / not included).
        materialization_records = []
        for result in results:
            if isinstance(result, VariantReadResult):
                if result.variant:  # Only include if variant is non-empty
                    rec = wasm_api_pb2.MaterializationRecord()
                    rec.unit = result.unit
                    rec.materialization = result.materialization
                    rec.rule = result.rule
                    rec.variant = result.variant
                    materialization_records.append(rec)
                # No prior assignment → omit (absence = no sticky assignment)
            elif isinstance(result, InclusionReadResult):
                if result.included:
                    rec = wasm_api_pb2.MaterializationRecord()
                    rec.unit = result.unit
                    rec.materialization = result.materialization
                    materialization_records.append(rec)
                # Not included → omit (absence = not included)

        return materialization_records

    def _write_materializations(
        self, records: List[wasm_api_pb2.MaterializationRecord]
    ) -> None:
        """Write materializations to the store.

        Args:
            records: The MaterializationRecords from a Resolved response.
        """
        try:
            ops = []
            for record in records:
                ops.append(
                    VariantWriteOp(
                        unit=record.unit,
                        materialization=record.materialization,
                        rule=record.rule,
                        variant=record.variant,
                    )
                )
            self._materialization_store.write(ops)
        except MaterializationNotSupportedError:
            logger.warning("Materialization write not supported")
        except Exception as e:
            logger.error("Failed to write materializations: %s", e)

    def _flush_assigned(self) -> None:
        """Flush assigned logs."""
        if self._resolver is None or self._flag_logger is None:
            return

        try:
            with self._resolver_lock:
                log_data = self._resolver.flush_assigned()
            if log_data:
                self._flag_logger.write(log_data)
        except Exception as e:
            logger.error("Failed to flush assigned logs: %s", e)

    def _start_background_threads(self) -> None:
        """Start background threads for state polling and log flushing."""
        self._shutdown_event.clear()

        # State polling thread
        self._state_thread = threading.Thread(
            target=self._state_poll_loop,
            daemon=True,
            name="confidence-state-poll",
        )
        self._state_thread.start()

        # Log flush thread
        self._log_thread = threading.Thread(
            target=self._log_flush_loop,
            daemon=True,
            name="confidence-log-flush",
        )
        self._log_thread.start()

    def _state_poll_loop(self) -> None:
        """Background loop for state polling."""
        # Use shorter retry interval when NOT_READY
        retry_interval = 1.0

        while True:
            # Use short interval if NOT_READY, normal interval otherwise
            interval = (
                retry_interval
                if self._status == ProviderStatus.NOT_READY
                else self._state_poll_interval
            )

            if self._shutdown_event.wait(timeout=interval):
                break

            try:
                state, account_id, changed = self._state_fetcher.fetch()
                if changed and account_id:
                    sdk = types_pb2.Sdk(
                        id=types_pb2.SdkId.SDK_ID_PYTHON_PROVIDER,
                        version=__version__,
                    )
                    with self._resolver_lock:
                        self._resolver.set_resolver_state(state, account_id, sdk)
                    logger.debug("Resolver state updated")

                # If we were NOT_READY and now have valid state, transition to READY
                if account_id and self._status == ProviderStatus.NOT_READY:
                    self._status = ProviderStatus.READY
                    self.emit_provider_ready(ProviderEventDetails())
                    logger.info("Provider recovered and is now READY")
            except Exception as e:
                logger.error("State fetch failed: %s", e)

    def _log_flush_loop(self) -> None:
        """Background loop for log flushing."""
        last_full_flush = 0.0
        last_assign_flush = 0.0

        while not self._shutdown_event.is_set():
            import time

            now = time.time()

            # Full flush at log_poll_interval
            if now - last_full_flush >= self._log_poll_interval:
                try:
                    with self._resolver_lock:
                        log_data = self._resolver.flush_logs()
                    if log_data and self._flag_logger is not None:
                        self._flag_logger.write(log_data)
                except Exception as e:
                    logger.error("Failed to flush logs: %s", e)
                last_full_flush = now

            # Assign flush at assign_poll_interval
            if now - last_assign_flush >= self._assign_poll_interval:
                self._flush_assigned()
                last_assign_flush = now

            # Sleep for shortest interval
            self._shutdown_event.wait(timeout=min(self._assign_poll_interval, 0.1))

    @staticmethod
    def _parse_flag_path(flag_key: str) -> Tuple[str, str]:
        """Parse a flag key into flag name and path.

        Args:
            flag_key: The flag key (e.g., "my-flag.nested.value").

        Returns:
            Tuple of (flag_name, path) where path may be empty.
        """
        parts = flag_key.split(".", 1)
        if len(parts) == 1:
            return parts[0], ""
        return parts[0], parts[1]

    @staticmethod
    def _context_to_proto(
        context: Optional[EvaluationContext],
    ) -> Optional[struct_pb2.Struct]:
        """Convert EvaluationContext to protobuf Struct.

        Args:
            context: The evaluation context.

        Returns:
            The protobuf Struct, or None if context is None.
        """
        if context is None:
            return None

        fields: Dict[str, struct_pb2.Value] = {}

        # Add targeting key as targeting_key
        if context.targeting_key:
            fields["targeting_key"] = struct_pb2.Value(
                string_value=context.targeting_key
            )

        # Add attributes
        if context.attributes:
            for key, value in context.attributes.items():
                fields[key] = ConfidenceProvider._value_to_proto(value)

        return struct_pb2.Struct(fields=fields)

    @staticmethod
    def _value_to_proto(value: Any) -> struct_pb2.Value:
        """Convert a Python value to protobuf Value.

        Args:
            value: The Python value.

        Returns:
            The protobuf Value.
        """
        if value is None:
            return struct_pb2.Value(null_value=struct_pb2.NullValue.NULL_VALUE)
        elif isinstance(value, bool):
            return struct_pb2.Value(bool_value=value)
        elif isinstance(value, (int, float)):
            return struct_pb2.Value(number_value=float(value))
        elif isinstance(value, str):
            return struct_pb2.Value(string_value=value)
        elif isinstance(value, list):
            list_value = struct_pb2.ListValue(
                values=[ConfidenceProvider._value_to_proto(v) for v in value]
            )
            return struct_pb2.Value(list_value=list_value)
        elif isinstance(value, dict):
            struct_value = struct_pb2.Struct(
                fields={
                    k: ConfidenceProvider._value_to_proto(v) for k, v in value.items()
                }
            )
            return struct_pb2.Value(struct_value=struct_value)
        else:
            return struct_pb2.Value(string_value=str(value))

    @staticmethod
    def _proto_struct_to_dict(struct: struct_pb2.Struct) -> Dict[str, Any]:
        """Convert protobuf Struct to Python dict.

        Args:
            struct: The protobuf Struct.

        Returns:
            The Python dict.
        """
        if struct is None:
            return {}

        result: Dict[str, Any] = {}
        for key, value in struct.fields.items():
            result[key] = ConfidenceProvider._proto_value_to_python(value)
        return result

    @staticmethod
    def _proto_value_to_python(value: struct_pb2.Value) -> Any:
        """Convert protobuf Value to Python value.

        Args:
            value: The protobuf Value.

        Returns:
            The Python value.
        """
        kind = value.WhichOneof("kind")
        if kind == "null_value":
            return None
        elif kind == "bool_value":
            return value.bool_value
        elif kind == "number_value":
            return value.number_value
        elif kind == "string_value":
            return value.string_value
        elif kind == "list_value":
            return [
                ConfidenceProvider._proto_value_to_python(v)
                for v in value.list_value.values
            ]
        elif kind == "struct_value":
            return ConfidenceProvider._proto_struct_to_dict(value.struct_value)
        else:
            return None

    @staticmethod
    def _get_value_for_path(path: str, value: Any) -> Tuple[Any, bool]:
        """Extract a nested value using dot notation.

        Args:
            path: The dot-separated path (e.g., "nested.value").
            value: The value to extract from.

        Returns:
            Tuple of (extracted_value, found).
        """
        if not path:
            return value, True

        parts = path.split(".")
        current = value

        for part in parts:
            if isinstance(current, dict) and part in current:
                current = current[part]
            else:
                return None, False

        return current, True

    def get_prometheus_metrics(self, request: Optional["SnapshotConfig"] = None) -> str:
        """Get a Prometheus metrics snapshot.

        **Experimental:** this API is subject to change.

        Args:
            request: Optional config (reserved for future options).

        Returns:
            The Prometheus metrics text, or empty string if not initialized.
        """
        if self._resolver is None:
            return ""
        with self._resolver_lock:
            return self._resolver.prometheus_snapshot()

    @staticmethod
    def _map_resolve_reason(reason: types_pb2.ResolveReason) -> Reason:
        """Map protobuf ResolveReason to OpenFeature Reason.

        Args:
            reason: The protobuf ResolveReason.

        Returns:
            The OpenFeature Reason.
        """
        if reason == types_pb2.ResolveReason.RESOLVE_REASON_MATCH:
            return Reason.TARGETING_MATCH
        elif reason == types_pb2.ResolveReason.RESOLVE_REASON_NO_SEGMENT_MATCH:
            return Reason.DEFAULT
        elif reason == types_pb2.ResolveReason.RESOLVE_REASON_FLAG_ARCHIVED:
            return Reason.DISABLED
        elif reason in (
            types_pb2.ResolveReason.RESOLVE_REASON_TARGETING_KEY_ERROR,
            types_pb2.ResolveReason.RESOLVE_REASON_ERROR,
            types_pb2.ResolveReason.RESOLVE_REASON_UNRECOGNIZED_TARGETING_RULE,
            types_pb2.ResolveReason.RESOLVE_REASON_MATERIALIZATION_NOT_SUPPORTED,
        ):
            return Reason.ERROR
        else:
            return Reason.UNKNOWN



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from . import internal_api_pb2 as confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2

GRPC_GENERATED_VERSION = '1.78.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + ' but the generated code in confidence/flags/resolver/v1/internal_api_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class InternalFlagLoggerServiceStub(object):
    """Simplified from confidence-resolver/protos/confidence/flags/resolver/v1/internal_api.proto
    Annotations removed for minimal dependencies

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ClientWriteFlagLogs = channel.unary_unary(
                '/confidence.flags.resolver.v1.InternalFlagLoggerService/ClientWriteFlagLogs',
                request_serializer=confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2.WriteFlagLogsRequest.SerializeToString,
                response_deserializer=confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2.WriteFlagLogsResponse.FromString,
                _registered_method=True)
        self.WriteMaterializedOperations = channel.unary_unary(
                '/confidence.flags.resolver.v1.InternalFlagLoggerService/WriteMaterializedOperations',
                request_serializer=confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2.WriteOperationsRequest.SerializeToString,
                response_deserializer=confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2.WriteOperationsResult.FromString,
                _registered_method=True)
        self.ReadMaterializedOperations = channel.unary_unary(
                '/confidence.flags.resolver.v1.InternalFlagLoggerService/ReadMaterializedOperations',
                request_serializer=confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2.ReadOperationsRequest.SerializeToString,
                response_deserializer=confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2.ReadOperationsResult.FromString,
                _registered_method=True)


class InternalFlagLoggerServiceServicer(object):
    """Simplified from confidence-resolver/protos/confidence/flags/resolver/v1/internal_api.proto
    Annotations removed for minimal dependencies

    """

    def ClientWriteFlagLogs(self, request, context):
        """Client writes flag assignment events and resolve logs using client secret authentication.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def WriteMaterializedOperations(self, request, context):
        """Stores materializations for units. Uses client secret authentication.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ReadMaterializedOperations(self, request, context):
        """Loads materializations for units. Uses client secret authentication.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_InternalFlagLoggerServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ClientWriteFlagLogs': grpc.unary_unary_rpc_method_handler(
                    servicer.ClientWriteFlagLogs,
                    request_deserializer=confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2.WriteFlagLogsRequest.FromString,
                    response_serializer=confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2.WriteFlagLogsResponse.SerializeToString,
            ),
            'WriteMaterializedOperations': grpc.unary_unary_rpc_method_handler(
                    servicer.WriteMaterializedOperations,
                    request_deserializer=confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2.WriteOperationsRequest.FromString,
                    response_serializer=confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2.WriteOperationsResult.SerializeToString,
            ),
            'ReadMaterializedOperations': grpc.unary_unary_rpc_method_handler(
                    servicer.ReadMaterializedOperations,
                    request_deserializer=confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2.ReadOperationsRequest.FromString,
                    response_serializer=confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2.ReadOperationsResult.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'confidence.flags.resolver.v1.InternalFlagLoggerService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('confidence.flags.resolver.v1.InternalFlagLoggerService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class InternalFlagLoggerService(object):
    """Simplified from confidence-resolver/protos/confidence/flags/resolver/v1/internal_api.proto
    Annotations removed for minimal dependencies

    """

    @staticmethod
    def ClientWriteFlagLogs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/confidence.flags.resolver.v1.InternalFlagLoggerService/ClientWriteFlagLogs',
            confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2.WriteFlagLogsRequest.SerializeToString,
            confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2.WriteFlagLogsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def WriteMaterializedOperations(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/confidence.flags.resolver.v1.InternalFlagLoggerService/WriteMaterializedOperations',
            confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2.WriteOperationsRequest.SerializeToString,
            confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2.WriteOperationsResult.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ReadMaterializedOperations(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/confidence.flags.resolver.v1.InternalFlagLoggerService/ReadMaterializedOperations',
            confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2.ReadOperationsRequest.SerializeToString,
            confidence_dot_flags_dot_resolver_dot_v1_dot_internal__api__pb2.ReadOperationsResult.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



"""WASM-based flag resolver using wasmtime.

This module provides the WasmResolver class that interfaces with the
Confidence flag resolver WASM module for local flag resolution.
"""

import itertools
from datetime import datetime
from typing import Optional

from google.protobuf import message as protobuf_message
from google.protobuf.timestamp_pb2 import Timestamp
from wasmtime import Config, Engine, Func, FuncType, Linker, Module, Store, ValType

from confidence.proto.confidence.wasm import messages_pb2
from confidence.proto.confidence.wasm import wasm_api_pb2

_instance_counter = itertools.count()


class WasmResolver:
    """WASM-based flag resolver.

    Interfaces with the Confidence WASM resolver module for local flag resolution.
    Uses wasmtime for WASM execution.

    Memory protocol: Data is exchanged with WASM using a 4-byte little-endian
    length prefix stored at ptr-4. The length includes the 4 prefix bytes.
    """

    def __init__(self, wasm_bytes: bytes) -> None:
        """Initialize the WASM resolver.

        Args:
            wasm_bytes: The compiled WASM binary bytes.
        """
        self._instance_id = str(next(_instance_counter))

        # Create WASM engine and store with caching enabled for faster startup
        config = Config()
        config.cache = True
        self._engine = Engine(config)
        self._store = Store(self._engine)

        # Compile the WASM module
        self._module = Module(self._engine, wasm_bytes)

        # Register host functions and instantiate
        self._register_host_functions()

        # Get exported functions
        exports = self._instance.exports(self._store)
        self._wasm_msg_alloc = exports["wasm_msg_alloc"]
        self._wasm_msg_free = exports["wasm_msg_free"]
        self._wasm_msg_guest_set_resolver_state = exports[
            "wasm_msg_guest_set_resolver_state"
        ]
        self._wasm_msg_guest_resolve_flags = exports["wasm_msg_guest_resolve_flags"]
        self._wasm_msg_guest_bounded_flush_logs = exports[
            "wasm_msg_guest_bounded_flush_logs"
        ]
        self._wasm_msg_guest_bounded_flush_assign = exports[
            "wasm_msg_guest_bounded_flush_assign"
        ]
        self._wasm_msg_guest_prometheus_snapshot = exports[
            "wasm_msg_guest_prometheus_snapshot"
        ]
        self._wasm_msg_guest_register_resolve = exports[
            "wasm_msg_guest_register_resolve"
        ]
        self._memory = exports["memory"]

    def _register_host_functions(self) -> None:
        """Register host functions that can be called from WASM."""

        def current_time(ptr: int) -> int:
            """Host function to return current timestamp."""
            try:
                # Create timestamp from current time
                now = datetime.now()
                timestamp = Timestamp()
                timestamp.FromDatetime(now)

                # Create response wrapper with timestamp data
                response = messages_pb2.Response()
                response.data = timestamp.SerializeToString()

                # Transfer response to WASM memory
                return self._transfer(response.SerializeToString())
            except Exception as e:
                # Return error response
                error_response = messages_pb2.Response()
                error_response.error = str(e)
                return self._transfer(error_response.SerializeToString())

        # Create function type: takes one i32 parameter, returns one i32
        func_type = FuncType([ValType.i32()], [ValType.i32()])
        host_func_time = Func(self._store, func_type, current_time)

        linker = Linker(self._store.engine)

        # Define the import with module and name
        linker.define(
            self._store, "wasm_msg", "wasm_msg_host_current_time", host_func_time
        )

        # Instantiate the module with imports
        self._instance = linker.instantiate(self._store, self._module)

    def set_resolver_state(
        self,
        state: bytes,
        account_id: str,
        sdk: Optional[protobuf_message.Message] = None,
    ) -> None:
        """Set the resolver state in the WASM module.

        Args:
            state: The serialized resolver state bytes.
            account_id: The account ID for the resolver.
            sdk: Optional SDK identifier and version.
        """
        request = messages_pb2.SetResolverStateRequest()
        request.state = state
        request.account_id = account_id
        if sdk is not None:
            request.sdk.CopyFrom(sdk)

        req_ptr = self._transfer_request(request)
        resp_ptr = self._wasm_msg_guest_set_resolver_state(self._store, req_ptr)

        if resp_ptr != 0:
            self._consume_response(resp_ptr, None)

    def resolve_process(
        self, request: wasm_api_pb2.ResolveProcessRequest
    ) -> wasm_api_pb2.ResolveProcessResponse:
        """Resolve flags using the WASM module.

        Args:
            request: The resolve process request.

        Returns:
            The resolve process response.
        """
        req_ptr = self._transfer_request(request)
        resp_ptr = self._wasm_msg_guest_resolve_flags(self._store, req_ptr)

        response = wasm_api_pb2.ResolveProcessResponse()
        self._consume_response(resp_ptr, response)
        return response

    def register_resolve(self, request: wasm_api_pb2.RegisterResolveRequest) -> None:
        """Register a resolve evaluation for telemetry."""
        req_ptr = self._transfer_request(request)
        resp_ptr = self._wasm_msg_guest_register_resolve(self._store, req_ptr)
        if resp_ptr != 0:
            self._consume_response(resp_ptr, None)

    def flush_logs(self) -> bytes:
        """Flush all pending logs from the WASM module.

        Returns:
            The serialized WriteFlagLogsRequest bytes.
        """
        # Pass 0 for unbounded flush
        resp_ptr = self._wasm_msg_guest_bounded_flush_logs(self._store, 0)

        # Read raw response data
        data = self._consume(resp_ptr)
        response = messages_pb2.Response()
        response.ParseFromString(data)

        if response.HasField("error") and response.error:
            raise RuntimeError(f"WASM error: {response.error}")

        return response.data if response.data else b""

    def flush_assigned(self) -> bytes:
        """Flush pending assignment logs from the WASM module.

        Returns:
            The serialized assignment data bytes.
        """
        # Pass 0 for unbounded flush
        resp_ptr = self._wasm_msg_guest_bounded_flush_assign(self._store, 0)

        # Read raw response data
        data = self._consume(resp_ptr)
        response = messages_pb2.Response()
        response.ParseFromString(data)

        if response.HasField("error") and response.error:
            raise RuntimeError(f"WASM error: {response.error}")

        return response.data if response.data else b""

    def prometheus_snapshot(self) -> str:
        """Get a Prometheus metrics snapshot from the WASM module.

        Returns:
            The Prometheus metrics text.
        """
        request = messages_pb2.PrometheusSnapshotRequest()
        request.instance = self._instance_id

        req_ptr = self._transfer_request(request)
        resp_ptr = self._wasm_msg_guest_prometheus_snapshot(self._store, req_ptr)

        response = messages_pb2.PrometheusSnapshotResponse()
        self._consume_response(resp_ptr, response)
        return response.text

    def _transfer_request(self, message: protobuf_message.Message) -> int:
        """Transfer a protobuf message to WASM memory as a Request.

        Args:
            message: The protobuf message to transfer.

        Returns:
            The pointer to the data in WASM memory.
        """
        data = message.SerializeToString()
        request = messages_pb2.Request()
        request.data = data
        return self._transfer(request.SerializeToString())

    def _transfer(self, data: bytes) -> int:
        """Allocate memory in WASM and copy data.

        Args:
            data: The bytes to copy to WASM memory.

        Returns:
            The pointer to the data in WASM memory.
        """
        # Allocate memory in WASM
        ptr = self._wasm_msg_alloc(self._store, len(data))

        # Write data to WASM memory
        self._memory.write(self._store, data, ptr)

        return ptr

    def _consume_response(
        self, addr: int, target: Optional[protobuf_message.Message]
    ) -> None:
        """Consume a response from WASM memory.

        Args:
            addr: The address in WASM memory.
            target: Optional protobuf message to parse the data into.

        Raises:
            RuntimeError: If the response contains an error.
        """
        data = self._consume(addr)
        response = messages_pb2.Response()
        response.ParseFromString(data)

        if response.HasField("error") and response.error:
            raise RuntimeError(f"WASM error: {response.error}")

        if target is not None and response.data:
            target.ParseFromString(response.data)

    def _consume(self, addr: int) -> bytes:
        """Read data from WASM memory and free it.

        Memory protocol: 4-byte little-endian length prefix at addr-4.
        The length value includes the 4 prefix bytes.

        Args:
            addr: The address in WASM memory.

        Returns:
            The bytes read from memory.
        """
        # Read length (4-byte little-endian prefix at addr-4)
        len_bytes = self._memory.read(self._store, addr - 4, addr)
        total_len = int.from_bytes(len_bytes, byteorder="little")
        length = total_len - 4

        # Read data
        data = self._memory.read(self._store, addr, addr + length)

        # Make a copy before freeing (defensive copy)
        data_copy = bytes(data)

        # Free memory
        self._wasm_msg_free(self._store, addr)

        return data_copy



"""Confidence OpenFeature provider for Python.

This package provides a local WASM-based flag resolver for Confidence feature flags,
implementing the OpenFeature provider interface.
"""

from confidence.version import __version__
from confidence.provider import ConfidenceProvider, SnapshotConfig
from confidence.materialization import (
    MaterializationStore,
    MaterializationNotSupportedError,
    RemoteMaterializationStore,
    UnsupportedMaterializationStore,
)

__all__ = [
    "__version__",
    "ConfidenceProvider",
    "SnapshotConfig",
    "MaterializationStore",
    "MaterializationNotSupportedError",
    "RemoteMaterializationStore",
    "UnsupportedMaterializationStore",
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: confidence/wasm/wasm_api.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    3,
    '',
    'confidence/wasm/wasm_api.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..flags.resolver.v1 import api_pb2 as confidence_dot_flags_dot_resolver_dot_v1_dot_api__pb2
from ..flags.resolver.v1 import types_pb2 as confidence_dot_flags_dot_resolver_dot_v1_dot_types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1e\x63onfidence/wasm/wasm_api.proto\x12\x1c\x63onfidence.flags.resolver.v1\x1a&confidence/flags/resolver/v1/api.proto\x1a(confidence/flags/resolver/v1/types.proto\"\xac\x05\n\x15ResolveProcessRequest\x12V\n\x19\x64\x65\x66\x65rred_materializations\x18\x01 \x01(\x0b\x32\x31.confidence.flags.resolver.v1.ResolveFlagsRequestH\x00\x12m\n\x17static_materializations\x18\x02 \x01(\x0b\x32J.confidence.flags.resolver.v1.ResolveProcessRequest.StaticMaterializationsH\x00\x12U\n\x18without_materializations\x18\x03 \x01(\x0b\x32\x31.confidence.flags.resolver.v1.ResolveFlagsRequestH\x00\x12L\n\x06resume\x18\x04 \x01(\x0b\x32:.confidence.flags.resolver.v1.ResolveProcessRequest.ResumeH\x00\x1a\xb3\x01\n\x16StaticMaterializations\x12J\n\x0fresolve_request\x18\x01 \x01(\x0b\x32\x31.confidence.flags.resolver.v1.ResolveFlagsRequest\x12M\n\x10materializations\x18\x02 \x03(\x0b\x32\x33.confidence.flags.resolver.v1.MaterializationRecord\x1a\x66\n\x06Resume\x12M\n\x10materializations\x18\x01 \x03(\x0b\x32\x33.confidence.flags.resolver.v1.MaterializationRecord\x12\r\n\x05state\x18\x02 \x01(\x0c\x42\t\n\x07resolve\"\xe8\x03\n\x16ResolveProcessResponse\x12Q\n\x08resolved\x18\x01 \x01(\x0b\x32=.confidence.flags.resolver.v1.ResolveProcessResponse.ResolvedH\x00\x12S\n\tsuspended\x18\x02 \x01(\x0b\x32>.confidence.flags.resolver.v1.ResolveProcessResponse.SuspendedH\x00\x1a\xa8\x01\n\x08Resolved\x12\x44\n\x08response\x18\x01 \x01(\x0b\x32\x32.confidence.flags.resolver.v1.ResolveFlagsResponse\x12V\n\x19materializations_to_write\x18\x02 \x03(\x0b\x32\x33.confidence.flags.resolver.v1.MaterializationRecord\x1aq\n\tSuspended\x12U\n\x18materializations_to_read\x18\x01 \x03(\x0b\x32\x33.confidence.flags.resolver.v1.MaterializationRecord\x12\r\n\x05state\x18\x02 \x01(\x0c\x42\x08\n\x06result\"]\n\x15MaterializationRecord\x12\x0c\n\x04unit\x18\x01 \x01(\t\x12\x17\n\x0fmaterialization\x18\x02 \x01(\t\x12\x0c\n\x04rule\x18\x03 \x01(\t\x12\x0f\n\x07variant\x18\x04 \x01(\t\"\x1d\n\nLogMessage\x12\x0f\n\x07message\x18\x01 \x01(\t\"i\n\x16RegisterResolveRequest\x12;\n\x06reason\x18\x01 \x01(\x0e\x32+.confidence.flags.resolver.v1.ResolveReason\x12\x12\n\nlatency_us\x18\x02 \x01(\rB\x9d\x01\n,com.spotify.confidence.sdk.flags.resolver.v1B\x0cWasmApiProtoP\x01Z]github.com/spotify/confidence-resolver/openfeature-provider/go/confidence/internal/proto/wasmb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'confidence.wasm.wasm_api_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n,com.spotify.confidence.sdk.flags.resolver.v1B\014WasmApiProtoP\001Z]github.com/spotify/confidence-resolver/openfeature-provider/go/confidence/internal/proto/wasm'
  _globals['_RESOLVEPROCESSREQUEST']._serialized_start=147
  _globals['_RESOLVEPROCESSREQUEST']._serialized_end=831
  _globals['_RESOLVEPROCESSREQUEST_STATICMATERIALIZATIONS']._serialized_start=537
  _globals['_RESOLVEPROCESSREQUEST_STATICMATERIALIZATIONS']._serialized_end=716
  _globals['_RESOLVEPROCESSREQUEST_RESUME']._serialized_start=718
  _globals['_RESOLVEPROCESSREQUEST_RESUME']._serialized_end=820
  _globals['_RESOLVEPROCESSRESPONSE']._serialized_start=834
  _globals['_RESOLVEPROCESSRESPONSE']._serialized_end=1322
  _globals['_RESOLVEPROCESSRESPONSE_RESOLVED']._serialized_start=1029
  _globals['_RESOLVEPROCESSRESPONSE_RESOLVED']._serialized_end=1197
  _globals['_RESOLVEPROCESSRESPONSE_SUSPENDED']._serialized_start=1199
  _globals['_RESOLVEPROCESSRESPONSE_SUSPENDED']._serialized_end=1312
  _globals['_MATERIALIZATIONRECORD']._serialized_start=1324
  _globals['_MATERIALIZATIONRECORD']._serialized_end=1417
  _globals['_LOGMESSAGE']._serialized_start=1419
  _globals['_LOGMESSAGE']._serialized_end=1448
  _globals['_REGISTERRESOLVEREQUEST']._serialized_start=1450
  _globals['_REGISTERRESOLVEREQUEST']._serialized_end=1555
# @@protoc_insertion_point(module_scope)
